"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.CompositeService = void 0;
var _util = require("util");
var _mergeStream = _interopRequireDefault(require("merge-stream"));
var _Service = require("./Service");
var _Logger = require("./Logger");
var _stream = require("./util/stream");
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
class CompositeService {
  stopping = false;
  constructor(config) {
    this.config = config;
    if (this.config.windowsCtrlCShutdown) {
      require("generate-ctrl-c-event"); // make sure this module loads before we even start
    }

    const outputStream = (0, _mergeStream.default)();
    outputStream.pipe(process.stdout);
    this.logger = new _Logger.Logger(this.config.logLevel);
    outputStream.add(this.logger.output);
    this.logger.log("debug", (0, _util.formatWithOptions)({
      depth: null
    }, "Config: %O", config));
    process.on("SIGINT", () => this.handleShutdownSignal(130, "SIGINT"));
    process.on("SIGTERM", () => this.handleShutdownSignal(143, "SIGTERM"));
    if (process.stdin.isTTY) {
      process.stdin.setRawMode(true);
      process.stdin.on("data", buffer => {
        if (buffer.toString("utf8") === "\u0003") {
          this.handleShutdownSignal(130, "ctrl+c");
        }
      });
    }
    this.services = Object.entries(this.config.services).map(([id, config]) => new _Service.Service(id, config, this.logger, this.handleFatalError.bind(this)));
    this.serviceMap = new Map(this.services.map(service => [service.id, service]));
    outputStream.add(this.services.map(({
      output,
      id
    }) => output.pipe((0, _stream.mapStreamLines)(line => `${id} | ${line}\n`))));
    this.logger.log("debug", "Starting composite service...");
    Promise.all(this.services.map(service => this.startService(service))).then(() => this.logger.log("debug", "Started composite service"));
  }
  async startService(service) {
    const dependencies = service.config.dependencies.map(id => this.serviceMap.get(id));
    await Promise.all(dependencies.map(service => this.startService(service)));
    if (this.stopping) {
      await never();
    }
    await service.start();
    if (this.stopping) {
      await never();
    }
  }
  handleFatalError(message) {
    this.logger.log("error", `Fatal error: ${message}`);
    if (!this.stopping) {
      this.stop(1);
    }
  }
  handleShutdownSignal(exitCode, description) {
    if (!this.stopping) {
      this.logger.log("info", `Received shutdown signal (${description})`);
      this.stop(exitCode);
    }
  }
  stop(exitCode) {
    if (this.stopping) return;
    this.stopping = true;
    this.logger.log("debug", "Stopping composite service...");
    if (this.config.windowsCtrlCShutdown) {
      require("generate-ctrl-c-event").generateCtrlCAsync().catch(error => this.logger.log("error", String(error)));
    }
    Promise.all(this.services.map(service => this.stopService(service))).then(() => this.logger.log("debug", "Stopped composite service"))
    // Wait one micro tick for output to flush
    .then(() => process.exit(exitCode));
  }
  async stopService(service) {
    if (this.config.gracefulShutdown) {
      const dependents = this.services.filter(({
        config
      }) => config.dependencies.includes(service.id));
      await Promise.all(dependents.map(service => this.stopService(service)));
    }
    await service.stop(this.config.windowsCtrlCShutdown);
  }
}
exports.CompositeService = CompositeService;
function never() {
  return new Promise(() => {});
}
//# sourceMappingURL=CompositeService.js.map