"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Service = void 0;
var _util = require("util");
var _stream = require("stream");
var _cloneableReadable = _interopRequireDefault(require("cloneable-readable"));
var _ServiceProcess = require("./ServiceProcess");
var _InternalError = require("./InternalError");
var _createReadyContext = require("./createReadyContext");
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
const delay = (0, _util.promisify)(setTimeout);
class Service {
  output = (0, _cloneableReadable.default)(new _stream.PassThrough({
    objectMode: true
  }));
  outputClone = this.output.clone();
  crashes = [];
  constructor(id, config, logger, handleFatalError) {
    this.id = id;
    this.config = config;
    this.logger = logger;
    this.handleFatalError = handleFatalError;
  }
  die(context, error) {
    this.handleFatalError(`${context}: ${(0, _util.inspect)(error)}`);
    return never();
  }
  start() {
    if (this.stopResult) {
      this.logger.log("error", new _InternalError.InternalError("Cannot start after stopping").stack);
      return this.startResult;
    }
    if (!this.startResult) {
      this.logger.log("debug", `Starting service '${this.id}'...`);
      this.defineReady();
      this.startResult = this.startProcess().then(() => this.ready).then(() => {
        this.logger.log("debug", `Started service '${this.id}'`);
      });
    }
    return this.startResult;
  }
  defineReady() {
    const ctx = (0, _createReadyContext.createReadyContext)(this.outputClone);
    this.ready = promiseTry(() => this.config.ready(ctx)).finally(() => this.outputClone.destroy()).catch(error => this.die(`In \`service.${this.id}.ready\``, error));
  }
  async startProcess() {
    const proc = new _ServiceProcess.ServiceProcess(this.id, this.config, this.logger, () => {
      proc.output.unpipe();
      if (!this.stopResult) {
        this.handleCrash(proc);
      }
    });
    this.process = proc;
    proc.output.pipe(this.output, {
      end: false
    });
    try {
      await this.process.started;
    } catch (error) {
      await this.die(`Spawning process for service '${this.id}'`, error);
    }
  }
  async handleCrash(proc) {
    this.logger.log("info", `Service '${this.id}' crashed`);
    const delayPromise = delay(this.config.minimumRestartDelay);
    const crash = {
      date: new Date(),
      logTail: proc.logTail
    };
    if (this.config.crashesLength > 0) {
      this.crashes.push(crash);
      if (this.crashes.length > this.config.crashesLength) {
        this.crashes.shift();
      }
    }
    const isServiceReady = await isResolved(this.ready);
    const ctx = {
      serviceId: this.id,
      isServiceReady,
      crash,
      crashes: this.crashes
    };
    try {
      await this.config.onCrash(ctx);
    } catch (error) {
      await this.die(`In \`service.${this.id}.onCrash\``, error);
    }
    await delayPromise;
    if (this.stopResult) {
      return;
    }
    this.logger.log("info", `Restarting service '${this.id}'`);
    await this.startProcess();
  }
  stop(windowsCtrlCShutdown) {
    if (!this.stopResult) {
      if (!this.process || !this.process.isRunning()) {
        this.stopResult = Promise.resolve();
      } else {
        this.logger.log("debug", `Stopping service '${this.id}'...`);
        this.stopResult = this.process.end(windowsCtrlCShutdown).then(() => {
          this.logger.log("debug", `Stopped service '${this.id}'`);
        });
      }
    }
    return this.stopResult;
  }
}
exports.Service = Service;
function promiseTry(fn) {
  try {
    return Promise.resolve(fn());
  } catch (error) {
    return Promise.reject(error);
  }
}
function isResolved(promise) {
  return Promise.race([promise.then(() => true, () => false), Promise.resolve().then(() => false)]);
}
function never() {
  return new Promise(() => {});
}
//# sourceMappingURL=Service.js.map