"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ConfigValidationError = void 0;
exports.validateAndNormalizeConfig = validateAndNormalizeConfig;
var t = _interopRequireWildcard(require("ts-interface-checker"));
function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }
function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }
function once(fn) {
  var result;
  return function () {
    return result || (result = fn());
  };
}
var typeSuite0 = once(function () {
  return {
    CompositeServiceConfig: t.iface([], {
      "logLevel": t.opt(t.union(t.lit("debug"), t.lit("info"), t.lit("error"))),
      "gracefulShutdown": t.opt("boolean"),
      "windowsCtrlCShutdown": t.opt("boolean"),
      "serviceDefaults": t.opt("ServiceConfig"),
      "services": t.iface([], {})
    })
  };
});
var typeSuite1 = once(function () {
  return {
    ServiceConfig: t.iface([], {
      "dependencies": t.opt(t.array("string")),
      "cwd": t.opt("string"),
      "command": t.opt(t.union("string", t.array("string"))),
      "env": t.opt(t.iface([], {})),
      "ready": t.opt(t.func("void", t.param("ctx", "ReadyContext"))),
      "forceKillTimeout": t.opt("number"),
      "onCrash": t.opt(t.func(t.union("void", "void"), t.param("ctx", "OnCrashContext"))),
      "crashesLength": t.opt("number"),
      "logTailLength": t.opt("number"),
      "minimumRestartDelay": t.opt("number")
    })
  };
});
function validateAndNormalizeConfig(config) {
  validateType("CompositeServiceConfig", "config", config);
  const {
    logLevel = "info"
  } = config;
  const windowsCtrlCShutdown = process.platform === "win32" && Boolean(config.windowsCtrlCShutdown);
  const gracefulShutdown = !windowsCtrlCShutdown && Boolean(config.gracefulShutdown);
  const {
    serviceDefaults = {}
  } = config;
  doExtraServiceConfigChecks("config.serviceDefaults", serviceDefaults);
  const truthyServiceEntries = Object.entries(config.services).filter(([, value]) => value);
  if (truthyServiceEntries.length === 0) {
    throw new ConfigValidationError("`config.services` has no entries");
  }
  const services = {};
  for (const [id, config] of truthyServiceEntries) {
    services[id] = processServiceConfig(id, config, serviceDefaults);
  }
  validateDependencyTree(services);
  return {
    logLevel,
    gracefulShutdown,
    windowsCtrlCShutdown,
    services
  };
}

/**
 * Structural validations for ServerConfig not covered by validateType/ts-interface-checker
 */
function doExtraServiceConfigChecks(path, config) {
  if (typeof config.command !== "undefined" && (Array.isArray(config.command) ? !config.command.length || !config.command[0].trim() : !config.command.trim())) {
    throw new ConfigValidationError(`\`${path}.command\` has no binary part`);
  }
}
const serviceBaseDefaults = {
  dependencies: [],
  cwd: ".",
  command: undefined,
  // no default command
  env: process.env,
  ready: () => Promise.resolve(),
  forceKillTimeout: 5000,
  onCrash: ctx => {
    if (!ctx.isServiceReady) throw new Error("Crashed before becoming ready");
  },
  crashesLength: 0,
  logTailLength: 0,
  minimumRestartDelay: 0
};
function processServiceConfig(id, config, defaults) {
  const path = `config.services.${id}`;
  validateType("ServiceConfig", path, config);
  doExtraServiceConfigChecks(path, config);
  const merged = {
    ...serviceBaseDefaults,
    ...removeUndefinedProperties(defaults),
    ...removeUndefinedProperties(config)
  };
  if (merged.command === undefined) {
    throw new ConfigValidationError(`\`${path}.command\` is not defined`);
  }
  return {
    ...merged,
    command: normalizeCommand(merged.command),
    env: normalizeEnv(merged.env)
  };
}
function removeUndefinedProperties(object) {
  const result = {
    ...object
  };
  for (const [key, value] of Object.entries(result)) {
    if (value === undefined) {
      delete result[key];
    }
  }
  return result;
}
function normalizeCommand(command) {
  return Array.isArray(command) ? command : command.split(/\s+/).filter(Boolean);
}
function normalizeEnv(env) {
  return Object.fromEntries(Object.entries(env).filter(([, value]) => value !== undefined).map(([key, value]) => [key, String(value)]));
}
function validateDependencyTree(services) {
  const serviceIds = Object.keys(services);
  for (const [serviceId, {
    dependencies
  }] of Object.entries(services)) {
    for (const dependency of dependencies) {
      if (!serviceIds.includes(dependency)) {
        throw new ConfigValidationError(`Service "${serviceId}" depends on unknown service "${dependency}"`);
      }
      if (services[dependency].ready === serviceBaseDefaults.ready) {
        throw new ConfigValidationError(`Service "${serviceId}" depends on service "${dependency}" which has no defined \`ready\` config`);
      }
    }
  }
  for (const serviceId of serviceIds) {
    validateNoCyclicDeps(serviceId, []);
  }
  function validateNoCyclicDeps(serviceId, path) {
    const isLooped = path.includes(serviceId);
    if (isLooped) {
      throw new ConfigValidationError(`Service "${serviceId}" has cyclic dependency ${path.slice(path.indexOf(serviceId)).concat(serviceId).join(" -> ")}`);
    }
    for (const dep of services[serviceId].dependencies) {
      validateNoCyclicDeps(dep, [...path, serviceId]);
    }
    return null;
  }
}
class ConfigValidationError extends Error {
  constructor(message) {
    super(message);
    Object.setPrototypeOf(this, ConfigValidationError.prototype);
  }
}
exports.ConfigValidationError = ConfigValidationError;
ConfigValidationError.prototype.name = ConfigValidationError.name;
const tsInterfaceBuilderOptions = {
  ignoreIndexSignature: true
};
const checkers = (0, t.createCheckers)({
  ...typeSuite0(),
  ...typeSuite1()
});
function validateType(typeName, reportedPath, value) {
  const checker = checkers[typeName];
  checker.setReportedPath(reportedPath);
  const error = checker.validate(value);
  if (error) {
    throw new ConfigValidationError(getErrorMessage(error[0]));
  }
}
function getErrorMessage(error) {
  return getErrorMessageLines(error).join("\n");
}
function getErrorMessageLines(error) {
  let result = [`\`${error.path}\` ${error.message}`];
  if (error.nested) {
    for (const nested of error.nested) {
      result = result.concat(getErrorMessageLines(nested).map(s => `    ${s}`));
    }
  }
  return result;
}
//# sourceMappingURL=validateAndNormalizeConfig.js.map